<?php

// current problems
// - attachments inserted into posts won't update download count / also may have issues with attachment moderation

if(!defined('IN_MYBB'))
	die('This file cannot be accessed directly.');

$plugins->add_hook('showthread_start', 'preparser_parser_mod_posts');


// clear parser cache if badwords/smilies/mycode is modified in some way
$plugins->add_hook('admin_config_badwords_add_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_badwords_delete_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_badwords_edit_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_smilies_add_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_smilies_edit_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_smilies_delete_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_smilies_add_multiple_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_smilies_mass_edit_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_mycode_toggle_status_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_mycode_add_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_mycode_edit_commit', 'preparser_clear_cache');
$plugins->add_hook('admin_config_mycode_delete_commit', 'preparser_clear_cache');
// change in dateformat, bburl?

$plugins->add_hook('admin_tools_recount_rebuild_start', 'preparser_admin_clearcache');

function preparser_info() {
	return array(
		'name'			=> 'Preparser Cache',
		'description'	=> 'A simple, dynamic pre-parser cache to speed up forum script processing.',
		'website'		=> 'http://mybbhacks.zingaburga.com/',
		'author'		=> 'ZiNgA BuRgA',
		'authorsite'	=> 'http://zingaburga.com/',
		'version'		=> '1.13',
		'compatibility'	=> '14*,15*,16*,17*,18*',
		'guid'			=> ''
	);
}

function preparser_is_installed() {
	return $GLOBALS['db']->table_exists('parser_cache');
}
function preparser_install() {
	global $db;
	switch($db->type) {
		case 'sqlite': case 'sqlite2': case 'sqlite3':
			$db->write_query('
				CREATE TABLE '.$db->table_prefix.'parser_cache (
					hash CHAR(20) PRIMARY KEY,
					dateline BIGINT(30) NOT NULL,
					msg TEXT NOT NULL
				)'.$db->build_create_table_collation().'
			');
			// hope SQLite uses BINARY collation (default)
			break;
		case 'pgsql':
			$db->write_query('
				CREATE TABLE '.$db->table_prefix.'parser_cache (
					hash BYTEA,
					dateline BIGINT NOT NULL,
					msg TEXT NOT NULL,
					PRIMARY KEY (hash)
				)'.$db->build_create_table_collation().'
			');
			break;
		default:
			$db->write_query('
				CREATE TABLE '.$db->table_prefix.'parser_cache (
					hash BINARY(20),
					dateline BIGINT(30) UNSIGNED NOT NULL,
					msg MEDIUMTEXT NOT NULL,
					PRIMARY KEY (hash),
					KEY (dateline)
				)'.$db->build_create_table_collation().'
			');
	}
}
function preparser_uninstall() {
	global $db;
	$db->write_query('DROP TABLE '.$db->table_prefix.'parser_cache');
}
function preparser_activate() {
	global $db, $cache;
	$db->insert_query('tasks', array(
		'title' => 'Pre-parser cache cleanup',
		'description' => 'Removes old cached parser entries.',
		'file' => 'preparser_cleanup',
		'minute' => '40',
		'hour' => '10',
		'day' => '*',
		'month' => '*',
		'weekday' => '*',
		'nextrun' => TIME_NOW + 86400*7,
		'lastrun' => 0,
		'enabled' => 1,
		'logging' => 1,
		'locked' => 0,
	));
	$cache->update_tasks();
}
function preparser_deactivate() {
	global $db, $cache;
	$db->delete_query('tasks', 'file=\'preparser_cleanup\'');
	$cache->update_tasks();
}

function preparser_parser_mod_posts() {
	preparser_parser_mod(true);
}

function preparser_parser_mod($attachments=false) {
	global $parser, $preparser_rndseed;
	if(!is_object($parser)) {
		require_once MYBB_ROOT.'inc/class_parser.php';
		$parser = new postParser;
	}
	
	$preparser_rndseed = md5(uniqid(mt_rand(), true));
	
	if($attachments) {
		global $templates;
		$tplcache = array('postbit_attachments_thumbnails_thumbnail', 'postbit_attachments_images_image', 'postbit_attachments_attachment');
		if(!$templates->cache['postbit_attachments_attachment'])
			$templates->cache(implode(',', $tplcache));
		
		foreach($tplcache as &$tpl)
			$templates->cache[$tpl] = '<!-- __PREPARSER_ATTACH_{$GLOBALS[\'preparser_current_hash\']}_{$attachment[\'aid\']}_'.$preparser_rndseed.' -->'.$templates->cache[$tpl].'<!-- __PREPARSER_ATTACH_END_'.$preparser_rndseed.' -->';
		unset($tplcache);
	}
	
	$GLOBALS['preparser_defer_parse'] = array();
	control_object($parser, '
		function parse_message($message, $options=array()) {
			if($this->preparser_output_mode)
				return parent::parse_message($message, $options);
			return preparser_parser_parse($message, $options, '.intval($attachments).');
		}
	');
	$parser->preparser_output_mode = false;
	// we work around issues such as vB Quote mod by using two functions
	$GLOBALS['plugins']->add_hook('pre_output_page', 'preparser_parser_mod_out', -9999);
	$GLOBALS['plugins']->add_hook('pre_output_page', 'preparser_parser_mod_store', 9999);
}
function preparser_parser_parse(&$message, &$options, $attachments) {
	$hash = sha1($message.'_'.serialize($options));
	$GLOBALS['preparser_current_hash'] = $hash;
	$GLOBALS['preparser_defer_parse'][$hash] = array($message, $options);
	// fix for attachment MyCode
	$attachcode = '';
	if($attachments && stripos($message, '[attachment=') !== false) {
		preg_match_all('~\[attachment=\d+\]~i', $message, $matches);
		$attachcode = implode('', $matches[0]);
	}
	return '<!-- __PREPARSER_ID_'.$hash.'__'.$GLOBALS['preparser_rndseed'].' -->'.$attachcode.'<!-- __PREPARSER_PARSE_END_'.$GLOBALS['preparser_rndseed'].' -->';
}
function preparser_parser_mod_out(&$page) {
	global $preparser_defer_parse, $preparser_rndseed, $db, $parser;
	if(empty($preparser_defer_parse)) return $page;
	
	$repl = array();
	$query = $db->simple_select('parser_cache', 'hash,msg', 'hash IN (\''.implode('\',\'', array_map('preparser_db_escape_binary', array_keys($preparser_defer_parse))).'\')');
	while($msg = $db->fetch_array($query)) {
		$hash = bin2hex($msg['hash']);
		unset($preparser_defer_parse[$hash]);
		$repl[$hash] = $msg['msg'];
	}
	// uncached entries
	if(!empty($preparser_defer_parse)) {
		// gather attachment code
		$attachcode_s = $attachcode_r = array();
		preg_match_all('~\<\!\-\- __PREPARSER_ATTACH_([a-fA-F0-9]{40})_(\d+)_'.$preparser_rndseed.' \-\-\>(.*?)\<\!\-\- __PREPARSER_ATTACH_END_'.$preparser_rndseed.' \-\-\>~s', $page, $matches, PREG_SET_ORDER);
		if(!empty($matches)) foreach($matches as &$match) {
			$attachcode_s[$match[1]][] = '[attachment='.$match[2].']';
			$attachcode_r[$match[1]][] = $match[3];
		}
		
		
		$parser->preparser_output_mode = true;
		// if running relative time, capture attempts to do this and revert to absolute times
		if($GLOBALS['mybb']->version_code >= 1700) {
			control_object($GLOBALS['templates'], '
				function get($title, $eslashes=1, $htmlcomments=1) {
					if($title == "mycode_quote_post" && $eslashes==1 && $htmlcomments==0) {
						$ret = parent::get($title, $eslashes, $htmlcomments);
						return \'".($span&&(int)$match[1]?(($span="<span> (".my_date($GLOBALS["mybb"]->settings["dateformat"], (int)$match[1], "", 0)." ".my_date($GLOBALS["mybb"]->settings["timeformat"], (int)$match[1], "", 0).")</span>")?"":""):"")."\'.$ret;
					}
					return parent::get($title, $eslashes, $htmlcomments);
				}
			');
		}
		
		//$cache_add = array();
		// detect if pre_output_page hooks are being added during parser (vB Quote does this)
		global $plugins;
		$current_hooks = $plugins->hooks['pre_output_page'];
		// some uncached entries, cache them
		foreach($preparser_defer_parse as $hash => &$msgopts) {
			$msg = $parser->parse_message($msgopts[0], $msgopts[1]);
			
			// do attachment fix
			if(!empty($attachcode_s[$hash]))
				$msg = str_ireplace($attachcode_s[$hash], $attachcode_r[$hash], $msg);
			
			$repl[$hash] = '<!-- __PREPARSER_ID_'.$hash.'_START__'.$preparser_rndseed.' -->'.$msg.'<!-- __PREPARSER_ID_'.$hash.'_END__'.$preparser_rndseed.' -->';
			/* $cache_add[] = array(
				'hash' => preparser_db_escape_binary($hash),
				'dateline' => TIME_NOW,
				'msg' => $db->escape_string($msg)
			); */
		}
		// re-use the hooks array
		foreach($current_hooks as $pri => &$hooks)
			if(is_array($hooks)) foreach($hooks as $k => &$hook) {
				unset($plugins->hooks['pre_output_page'][$pri][$k]);
				if(empty($plugins->hooks['pre_output_page'][$pri]))
					unset($plugins->hooks['pre_output_page'][$pri]);
			}
		$rerun_hooks = !empty($plugins->hooks['pre_output_page']);
		//$db->insert_query_multiple('parser_cache', $cache_add);
	}
	
	
	
	// remove any hanging attachment comments (possible if they don't go thru [attachment] code)
	$pfind = array('~\<\!\-\- __PREPARSER_ATTACH_[0-9a-fA-F]{40}_\d+_'.$preparser_rndseed.' \-\-\>~s', '~\<\!\-\- __PREPARSER_ATTACH_END_'.$preparser_rndseed.' \-\-\>~');
	$prepl = array('', '');
	foreach($repl as $hash => &$rpl) {
		$pfind[] = '~\<\!\-\- __PREPARSER_ID_'.$hash.'__'.$preparser_rndseed.' \-\-\>.*?\<\!\-\- __PREPARSER_PARSE_END_'.$preparser_rndseed.' \-\-\>~s';
		$prepl[] = strtr($rpl, array('\\' => '\\\\', '$' => '\\$'));
	}
	$newpage = preg_replace($pfind, $prepl, $page);
	if($rerun_hooks)
		return $plugins->run_hooks('pre_output_page', $newpage);
	else
		return $newpage;
}

function preparser_parser_mod_store(&$page) {
	global $preparser_defer_parse, $preparser_rndseed, $db, $parser;
	if(empty($preparser_defer_parse)) return $page;
	$regex = '~\<\!\-\- __PREPARSER_ID_([a-fA-F0-9]{40})_START__'.$preparser_rndseed.' \-\-\>(.*?)\<\!\-\- __PREPARSER_ID_\\1_END__'.$preparser_rndseed.' \-\-\>~s';
	preg_match_all($regex, $page, $matches, PREG_SET_ORDER);
	
	if(empty($matches)) return $page;
	
	$msgs = array();
	foreach($matches as &$match) {
		// this will also take care of the case when there are multiple posts with the exact same content
		$msgs[$match[1]] = $match[2];
	}
	unset($matches);
	
	$cache_add = array();
	foreach($preparser_defer_parse as $hash => &$msgopts) {
		if(!isset($msgs[$hash])) continue;
		$cache_add[] = array(
			'hash' => preparser_db_escape_binary($hash),
			'dateline' => TIME_NOW,
			'msg' => $db->escape_string($msgs[$hash])
		);
	}
	if(!empty($cache_add)) {
		// this may very well introduce a race condition - so suppress errors if this is the case
		//$db->insert_query_multiple('parser_cache', $cache_add);
		$vals = '';
		foreach($cache_add as &$c)
			$vals .= ($vals?',':'').'(\''.implode('\',\'', $c).'\')';
		$db->write_query('INSERT '.($db->type == 'mysql' || $db->type == 'mysqli' ? 'IGNORE ':'').'INTO '.$db->table_prefix.'parser_cache (hash,dateline,msg) VALUES'.$vals, true);
	}
	
	// not the fastest method, but as this function, hopefully, isn't called often, it shouldn't be much of an issue
	return preg_replace($regex, '$2', $page);
}

function preparser_clear_cache() {
	global $db;
	$db->write_query('TRUNCATE TABLE '.$db->table_prefix.'parser_cache');
}



function preparser_admin_clearcache() {
	global $mybb, $db;
	if($mybb->request_method == 'post') {
		if(isset($mybb->input['do_clearparsercache'])) {
			preparser_clear_cache();
			flash_message('Parser cache cleared', 'success');
			admin_redirect('index.php?module=tools'.($mybb->version_code>=1500?'-':'/').'recount_rebuild');
		}
	}
	else {
		$GLOBALS['plugins']->add_hook('admin_formcontainer_end', 'preparser_admin_clearcache_show');
	}
}
function preparser_admin_clearcache_show() {
	global $form_container, $form, $lang;
	
	$form_container->output_cell('<label>Clear Parser Cache</label><div class="description">Removes all preparsed posts/signatures/whatever from cache (these will automatically be rebuilt on the fly).</div>');
	$form_container->output_cell($lang->na);
	$form_container->output_cell($form->generate_submit_button($lang->go, array('name' => 'do_clearparsercache')));
	$form_container->construct_row();
}



function preparser_db_escape_binary($s) {
	global $db;
	$s = pack('H*', $s);
	if($db->type == 'pgsql')
		return pg_escape_bytea($db->current_link, $s);
	
	if(isset($db->db_encoding)) { // hack for MyBB >= 1.6.12 to force it to not screw up our binary field
		$old_db_encoding = $db->db_encoding;
		$db->db_encoding = 'binary';
	}
	$ret = $db->escape_string($s);
	if(isset($old_db_encoding)) $db->db_encoding = $old_db_encoding;
	return $ret;
}

if(!function_exists('control_object')) {
	function control_object(&$obj, $code) {
		static $cnt = 0;
		$newname = '_objcont_'.(++$cnt);
		$objserial = serialize($obj);
		$classname = get_class($obj);
		$checkstr = 'O:'.strlen($classname).':"'.$classname.'":';
		$checkstr_len = strlen($checkstr);
		if(substr($objserial, 0, $checkstr_len) == $checkstr) {
			$vars = array();
			// grab resources/object etc, stripping scope info from keys
			foreach((array)$obj as $k => $v) {
				if($p = strrpos($k, "\0"))
					$k = substr($k, $p+1);
				$vars[$k] = $v;
			}
			if(!empty($vars))
				$code .= '
					function ___setvars(&$a) {
						foreach($a as $k => &$v)
							$this->$k = $v;
					}
				';
			eval('class '.$newname.' extends '.$classname.' {'.$code.'}');
			$obj = unserialize('O:'.strlen($newname).':"'.$newname.'":'.substr($objserial, $checkstr_len));
			if(!empty($vars))
				$obj->___setvars($vars);
		}
		// else not a valid object or PHP serialize has changed
	}
}
